

'-------------------------------------------------------------------------------
'
'' Module: i2c_int_keyboard.bas
' *** Aangepast voor Mahjong testsetup
'
' Copyright 2005-2008 Elba Corp.
' All rights reserved.
'
' Non-exclusive permission to use, copy, modify, and distribute this
' software for any purpose and without fee is hereby granted, provided:
'   1) that the above copyright notice is retained in the source code and,
'   2) that it is understood and agreed that no warranty of any kind,
'      express or implied is provided.
' Use of this code in any manner shall be prima facie evidence of
' acceptance of these terms.
'
' You may contact the author by email at dkinzer@zbasic.net.
'
'-------------------------------------------------------------------------------
'
' Overview:
'
' Matrix keypads are relatively inexpensive and serve well to facilitate
' human input to an embedded system.  A typical embedded system keypad might
' have 12 keys arranged in a 4 row by 3 column matrix.  Each keyswitch connects
' one row line to one column lead.
'
' To interface such a keypad to a microcontroller, it is necessary to have a
' means to activate one row at a time and then examine the column output to
' determine if a key in that row is pressed.  Alternately, the column lines
' can be driven and the row lines observed.
'
' The routines in this module implement the scanning process using a PCF8574
' I/O Expander chip on the I2C bus.  The PCF8574A could be used; the only
' difference between the two chips the the I2C slave address.
'
' This code contains operates the keypad in an interrupt-driven mode, utilizing
' the /INT output of the I/O expander and the WaitForInterrupt() routine of ZBasic.
'
' Lastly, this module contains a simple Main() driver routine intended to be
' used for testing the code.  The Main() is contained within a conditional
' construct so that it can easily be enabled or disabled as necessary.
'
' This code is intended to accompany application note AN205 which contains
' additional information on the hardware connections.
'
'-------------------------------------------------------------------------------

'Resources
Public keyPadTaskStack(1 to 114) as Byte
Public keyQueue(1 to 20) as Byte

' define the keyboard layout
Private Const kbdRows as Byte = 4    ' number of keypad rows
Private Const kbdCols as Byte = 3    ' number of keypad columns

' delay times, in RTC ticks
Private Const debounceDelay as UnsignedInteger = 15
Private Const autoRepeatDelay As Long = 90
Private Const keyWait As UnsignedInteger = 25

' define some special keys
Private Const KeyEnter as Byte = &H23
Private Const KeyDecimal as Byte = &H2a

' uncomment the line below to use the I2C hardware channel
#define USE_HW_CHAN

#ifdef USE_HW_CHAN
Private Const i2cChan as Byte = 0
#else
Private Const i2cChan as Byte = 1

' These lines define the pin numbers used for the SDA and SCL signals.
' They are needed only if the hardware channel is not used.
#if Option.TargetDevice = "ZX1281" OR Option.TargetDevice = "ZX1280"
Private Const sclPin as Byte = D.0	' to pin 14 of PCF8574A
Private Const sdaPin as Byte = D.1	' to pin 15 of PCF8574A
#else
Private Const sclPin as Byte = C.0	' to pin 14 of PCF8574A
Private Const sdaPin as Byte = C.1	' to pin 15 of PCF8574A
#endif
#endif

' define the address components of the PCF8574A
Private Const i2cAddrBase as Byte = &H70	' this part is fixed
Private Const i2cAddrOfst as Byte = &H00	' this part is selectable by A2-A0
Private Const i2cAddr as Byte = i2cAddrBase + i2cAddrOfst


'-------------------------------------------------------------------------------
'
'' setRowBits
'
' Set the state of the row lines.
'
Private Sub setRowBits(ByVal rowBits as Byte)
	Dim ival as Integer
	
	' set all column bits high, row bits as specified
	rowBits = rowBits Or &Hf0
	ival = I2CCmd(i2cChan, i2cAddr, 1, rowBits, 0, 0)
End Sub

'-------------------------------------------------------------------------------
'
'' KeyboardInit
'
' Initialize the keyboard interface circuitry.
'
Private Sub KeyboardInit()
	Dim data as Byte
	Dim ival as Integer

#ifdef USE_HW_CHAN
	Call OpenI2C(i2cChan, 0, 0)
#else
	Call OpenI2C(i2cChan, sdaPin, sclPin)
#endif

	' set all outputs high
	Call setRowBits(&Hff)
End Sub

'-------------------------------------------------------------------------------
'
'' KeyIsDown
'
' Determine if a key is pressed.  If so, return its key code, otherwise, 0.
' Note that this function returns the code of the first key detected.  No
' effort is made to determine if multiple keys are down.
'
Private Function KeyIsDown() as Byte
	Dim row as Byte, col as Byte
	Dim rowBits as Byte, colBits as Byte

	KeyIsDown = 0

	' scan by rows, looking for a key that is down
	rowBits = &Hfe
 	For row = 1 to kbdRows
		Dim ival as Integer

		' set one row line low, read back the column data
		ival = I2CCmd(i2cChan, i2cAddr, 1, rowBits, 1, colBits)

		' scan the column bits looking for a zero
		Const colMask as Byte = Shl(1, kbdRows)
		Dim colBitsSave as Byte
		colBitsSave = colBits
		For col = 1 to kbdCols
			If (Not CBool(colBits And colMask)) Then
 				KeyIsDown = keyCode(row, col)
 				Exit Function
 			End If
			colBits = Shr(colBits, 1)
 		Next
 
 		' prepare to scan the next row
		rowBits = Shl(rowBits, 1) Or &H01
 	Next
End Function

'-------------------------------------------------------------------------------

' This data table maps row/column pairs to ASCII codes.
' Note that it must be indexed as keyTable(col, row).
Private keyTable as ByteTableData({
	"123A"
	"456B"
	"789C"
	"*0#D"
})

'-------------------------------------------------------------------------------
'
'' keyCode
'
' Given row and column values, return the code for the corresponding
' key or zero if the row/column combination is invalid.
'
Private Function keyCode(ByVal row as Byte, ByVal col as Byte) as Byte
	If ((row >= 1) And (row <= CByte(UBound(keyTable, 2))) And _
			(col >= 1) And (col <= CByte(UBound(keyTable, 1)))) Then
		keyCode = keyTable(col, row)
	Else
		keyCode = 0
	End If
End Function

'-------------------------------------------------------------------------------

'
'' KeyPadTask
'
' This task waits for a key-down interrupt from the keypad and then reads
' the key, placing the key value in a queue.  Auto-repeat is implemented.
' Note, however, that N-key rollover is not implemented.  Each key must
' be pressed and then released before the next key may be pressed.
'
Public Sub KeyPadTask()
	Dim data as Byte
	
	' initialize the keypad interface
	Call KeyboardInit()
	Do
		Dim key as Byte

		' wait for a keypress interrupt
		Call setRowBits(&Hf0)
		Call WaitForInterrupt(zxPinFallingEdge, 1)

		' read the key, add to a queue
		Call setRowBits(&Hff)
		key = KeyIsDown()
		If (key <> 0) Then
			' debounce the key press
			Call Sleep(debounceDelay)
			
			' add the key to the queue
			Call PutQueue(keyQueue, key, 1)

			' auto-repeat
			Dim repeatTime as Long
			repeatTime = Register.RTCTick + autoRepeatDelay
			Do

				' wait a bit
				Call Sleep(keyWait)

				If (KeyIsDown() <> key) Then
					Exit Do
				End If

				' the key is still down, see if it is time to auto-repeate
				If (Register.RTCTick > repeatTime) Then
					' add the key to the queue again
					Call PutQueue(keyQueue, key, 1)

					' reset the threshold for another auto-repeat
					repeatTime = Register.RTCTick + autoRepeatDelay
				End If
			Loop

			' debounce the key release
			Call Sleep(debounceDelay)
		End If
	Loop
End Sub

'-------------------------------------------------------------------------------

' uncomment the next line to include a simple test driver
'#define KEYBOARD_TEST_
#ifdef KEYBOARD_TEST_

Sub Main()
	Call OpenQueue(keyQueue, SizeOf(keyQueue))
	CallTask keyPadTask, KeyPadTaskStack

	Do
		Do While (StatusQueue(keyQueue))
			Dim key as Byte
			Call GetQueue(keyQueue, key, 1)
			Debug.Print Chr(key)
		Loop
		Call Sleep(0.1)
	Loop
End Sub

#endif

